//___________________________________________________________________
//
//                           Class ListBox
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/textctrl.h>
#include <wx/stattext.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/glcanvas.h>
#include <wx/listbox.h>

#include "Main.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "Point.h"
#include "Error.h"
#include "Eval.h"
#include "Stack.h"

#include "Event.h"
#include "ListBox.h"
#include "Frame.h"

#include "PtrToInt.h"
#include "IntToPtr.h"

#include "Init.h"
#include "Color.h"
#include "Graph.h"
#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//___________________________________________________________________

class MyListBox : public wxListBox
{
public:
	MyListBox(wxWindow* parent, wxWindowID id, 
		const wxArrayString& choices, 
		const wxPoint& pos=wxDefaultPosition, 
		const wxSize& size=wxDefaultSize, 
		long style = 0, 
		const wxValidator& validator = wxDefaultValidator, 
		const wxString& name = "ListBox");

	// MSW sends extra kill focus event
	virtual ~MyListBox()
	{
		if (parent) 
			parent->listBox = 0;
		parent=0;
	}

	ListBox_ *parent;

	void OnChar(wxKeyEvent &event);         // pass chars to ListBox
	void OnKeyDown(wxKeyEvent &event);
	void OnKeyUp(wxKeyEvent &event);
	void OnSelect(wxCommandEvent &event);
	void OnDClick(wxCommandEvent &event);

private:
	DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(MyListBox,wxListBox)
	EVT_CHAR		(MyListBox::OnChar		)
	EVT_KEY_DOWN(MyListBox::OnKeyDown	)
	EVT_KEY_UP	(MyListBox::OnKeyUp		)
	EVT_LISTBOX			(-1,	MyListBox::OnSelect	)
	EVT_LISTBOX_DCLICK(-1,	MyListBox::OnDClick	)
END_EVENT_TABLE()

MyListBox::MyListBox(wxWindow *parent_, wxWindowID id,
							  const wxArrayString &choices_,
							  const wxPoint &pos, const wxSize &size,
							  long style,
							  const wxValidator& validator,
							  const wxString &name):
wxListBox(parent_, id, pos, size, choices_, style, validator, name)
{
	parent = (ListBox_*)parent_;
}

void MyListBox::OnChar(wxKeyEvent &event)
{
	if (parent) 
		parent->OnChar(event);
	event.Skip();
}

void MyListBox::OnKeyDown(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyDown(event);
	event.Skip();
}

void MyListBox::OnKeyUp(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyUp(event);
	event.Skip();
}

void MyListBox::OnSelect(wxCommandEvent &event)
{
	if (parent) 
		parent->OnSelect(event);
}

void MyListBox::OnDClick(wxCommandEvent &event)
{
	if (parent) 
		parent->OnDClick(event);
}
//___________________________________________________________________

BEGIN_EVENT_TABLE(ListBox_, wxControl)
	EVT_CHAR				(		ListBox_::OnChar		)
	EVT_KEY_DOWN		(		ListBox_::OnKeyDown	)
	EVT_KEY_UP			(		ListBox_::OnKeyUp		)
	EVT_LISTBOX			(-1,	ListBox_::OnSelect		)
	EVT_LISTBOX_DCLICK(-1,	ListBox_::OnDClick		)
END_EVENT_TABLE()

ListBox_::ListBox_(ListBoxNodeP listBoxNode_, int style):
label(0), listBox(0), bestSize(-999, -999),
choices(0),
onSelect(0), onDClick(0),
onChar(0), onKeyUp(0), onKeyDown(0), 
listBoxNode(listBoxNode_)
{
	wxWindow *parent=console;
	if (listBoxNode->parentFrame)
		parent=listBoxNode->parentFrame->frame;
	else if (listBoxNode->parentGraph)
		parent=listBoxNode->parentGraph->graph;
	create(parent, -1, wxArrayString(), 
			wxDefaultPosition, wxDefaultSize, style,
			wxDefaultValidator, "ListBox");
}

ListBox_::ListBox_(ListBoxNodeP listBoxNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxArrayString &choices_, 
			PairP onSelect_,
			PairP onDClick_,
			const wxPoint& pos, 
			const wxSize& size, 

			long style, 
			const wxValidator& validator, 
			const wxString& name):
label(0), listBox(0), bestSize(-999, -999), choices(choices_),
onSelect(onSelect_), onDClick(onDClick_),
onChar(0), onKeyUp(0), onKeyDown(0), 
listBoxNode(listBoxNode_)
{
	bestSize = wxSize(-999, -999);
	listBoxNode->listBox=this;
	listBoxNode->style=style;
	create(parent, id, choices, pos, size, style, validator, name);
}

bool ListBox_::create(wxWindow* parent, wxWindowID id, 
			const wxArrayString &choices_, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name)
{
	if (!wxControl::Create(parent, id, pos, size, style|wxNO_BORDER,
								wxDefaultValidator, name))
	  return false;

	wxControl::SetLabel(name);
	wxControl::SetBackgroundColour(parent->GetBackgroundColour());
	wxControl::SetForegroundColour(parent->GetForegroundColour());

	int width = size.GetWidth(), height = size.GetHeight();

	wxSize best_size(DoGetBestSize());
	if (width  == -1) width  = best_size.GetWidth();
	if (height == -1) height = best_size.GetHeight();

	if (name != wxEmptyString)
		label = new wxStaticText(this, id, name, 
			wxPoint(0, 0), wxDefaultSize, 0, "staticText");
	else
		label=0;

	if (label)
		listBox = new MyListBox(this, id, choices_,
			wxPoint(0,0),
			wxSize(width, height-label->GetSize().GetHeight()),
			style, validator, name);
	else
		listBox = new MyListBox(this, id, choices_,
			wxPoint(0,0),
			wxSize(width, height), style, validator, name);

	choices.Clear();

	DoSetSize(pos.x, pos.y, width, height);
	SetBestFittingSize(wxSize(width, height));
	return true;
}

void ListBox_::DoSetSize(int x, int y, int width, int height, int sizeFlags)
{
	wxSize bestSize(DoGetBestSize());
	if (width < 0)  width  = bestSize.GetWidth();
	if (height < 0) height = bestSize.GetHeight();

	wxWindow::DoSetSize(x, y, width, height, sizeFlags);

	int labelwidth  = 0;
	int labelheight = 0;
	if (label)
		label->GetSize(&labelwidth, &labelheight);

	if (listBox)   
		listBox->SetSize(0, labelheight, 
			width, height-labelheight-2);
}

wxSize ListBox_::DoGetBestSize()
{
	if (bestSize.x == -999)
	{	wxListBox *t=new wxListBox((wxWindow*)this, wxID_ANY, 
			wxPoint(0,0), wxDefaultSize, choices, GetWindowStyle());
		bestSize=t->GetBestSize();
		delete t;
		if (GetLabel()!=wxEmptyString)
		{	wxStaticText label((wxWindow*)this, wxID_ANY, GetLabel());
			wxSize ls=label.GetBestSize();
			bestSize.y+=ls.y;
			if (ls.x > bestSize.x)
				bestSize.x=ls.x;
		}
	}
	return bestSize;
}

ListBox_::~ListBox_()
{
	if (listBoxNode)
		listBoxNode->listBox=0;
	if (label)
	{	delete label;
		label=0;
	}
	if (listBox)
	{	delete listBox;
		listBox=0;
	}

	EventPP out=Event::fifo.out;
	EventP ev;
	while (Event::fifo)
	{	Event::fifo >> ev;
		switch (ev->type)
		{
		case Event::ListBoxSelect:
		case Event::ListBoxDClick:
		case Event::ListBoxChar:
		case Event::ListBoxKeyDown:
		case Event::ListBoxKeyUp:
			if (((EventListBox*)ev)->listBox==this)
				((EventListBox*)ev)->listBox=0;
			break;
		default:
			;
		}
	}
	Event::fifo.out=out;
}

void ListBox_::Set(wxArrayString &choices)
{
	if (listBox)
		listBox->Set(choices);
}

#define LB_ON(NAME)																\
void ListBox_::On##NAME(wxKeyEvent &event)									\
{																						\
	Event::fifo << new EventListBox##NAME(this, event.GetKeyCode());	\
	event.Skip();																	\
}

LB_ON(Char)
LB_ON(KeyDown)
LB_ON(KeyUp)

void ListBox_::OnSelect(wxCommandEvent &event)
{
	Event::fifo << new EventListBoxSelect(this);
}

void ListBox_::OnDClick(wxCommandEvent &event)
{
	Event::fifo << new EventListBoxDClick(this);
}

wxString ListBox_::wxstring()
{
	wxString str;
	str << "ListBox " 
		<< '[' << GetName() << "] ";
	if (onSelect)
		str << onSelect->string()->ptr() << ' ';
	else
		str << "[] ";
	if (onDClick)
		str << onDClick->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

wxDataInputStream &ListBox_::read(wxDataInputStream &istr) 
{	
	wxPoint pos;
	wxSize size;
	wxString name, fontName;
	int fontSize, fontStyle, fontWeight;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> name 
		>> (int&)onSelect 
		>> (int&)onDClick
		>> (int&)onChar
		>> (int&)onKeyDown
		>> (int&)onKeyUp
		>> pos.x >> pos.y >> size.x >> size.y
		>> fr >> fg >> fb 
		>> br >> bg >> bb 
		>> fontSize >> fontName >> fontStyle >> fontWeight;

	int nr, i;
	istr >> nr;
	choices.Clear();
	for (i=0; i<nr; i++)
	{	wxString item;
		istr >> item;
		choices.Add(item);
	}
	listBox->Set(choices);

	istr >> nr;
	for (i=0; i<nr; i++)
	{	int sel;
		istr >> sel;
		listBox->SetSelection(sel);
	}
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	wxFont f=GetFont();
	f.SetPointSize(fontSize);
	f.SetFaceName(fontName);
	f.SetStyle(fontStyle);
	f.SetWeight(fontWeight);
	SetFont(f);
	return istr;
}
wxDataOutputStream &ListBox_::write(wxDataOutputStream &ostr) 
{
//	flush();
	ostr
		<< GetName() 
		<< GC_::ptrToInt->find(onSelect)
		<< GC_::ptrToInt->find(onDClick)
		<< GC_::ptrToInt->find(onChar)
		<< GC_::ptrToInt->find(onKeyDown)
		<< GC_::ptrToInt->find(onKeyUp)
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y;

	{	wxColour 
			f=listBox->GetForegroundColour(), 
			b=listBox->GetBackgroundColour();
		ostr
			<< f.Red() << f.Green() << f.Blue() 
			<< b.Red() << b.Green() << b.Blue();
	}
	wxFont f=listBox->GetFont();
	ostr 
		<< f.GetPointSize() << f.GetFaceName() 
		<< f.GetStyle() << f.GetWeight();

	int nr=listBox->GetCount(), i;
	ostr << nr;
	for (i=0; i<nr; i++)
		ostr << listBox->GetString(i);

	wxArrayInt sel;
	nr=listBox->GetSelections(sel);
	ostr << nr;
	for (i=0; i<nr; i++)
		ostr << sel[i];
	return ostr;
}

bool ListBox_::SetBackgroundColour(const wxColour &c)
{
	if (listBox)
		return listBox->SetBackgroundColour(c);
	return false;
}

bool ListBox_::SetForegroundColour(const wxColour &c)
{
	if (listBox)
		return listBox->SetForegroundColour(c);
	return false;
}

bool ListBox_::SetFont(const wxFont &f)
{
	if (listBox)
		return listBox->SetFont(f);
	return false;
}

//___________________________________________________________________

#define NODE_ON(x)							\
PairP ListBoxNode::on##x()					\
{													\
	if (listBox)								\
		return listBox->on##x;				\
	return 0;									\
}													\
													\
void ListBoxNode::seton##x(NodeP nd)	\
{													\
	if (listBox)								\
	{	listBox->on##x = PairP(nd);		\
		check_oldyoung(nd);					\
	}												\
}

NODE_ON(Char)
NODE_ON(KeyDown)
NODE_ON(KeyUp)
NODE_ON(Select)
NODE_ON(DClick)

void ListBoxNode::setParentFrame(FrameNodeP frame)
{
	parentFrame=frame;
}

void ListBoxNode::setParentGraph(GraphNodeP graph)
{
	parentGraph=graph;
}

ListBoxNode::ListBoxNode(): 
	Node(LISTBOX), listBox(0), 
	parentFrame(0), 
	parentGraph(0) 
{}

ListBoxNode::ListBoxNode(FrameNodeP parentFrame_, 
											GraphNodeP parentGraph_): 
	Node(LISTBOX), listBox(0), 
	parentFrame(parentFrame_),
	parentGraph(parentGraph_) 
{
	FrameNodeP pf=parentFrame;
	GraphNodeP pg=parentGraph;
	if (pf)
	{	pf->children = pf->children->wlPut(this);
		pf->check_oldyoung(pf->children);
	}
	else if (pg)
	{	pg->children = pg->children->wlPut(this);
		pg->check_oldyoung(pg->children);
	}
}

void ListBoxNode::free() 
{	
	if (listBox) 
	{	listBox->listBoxNode=0;
		listBox->Destroy(); 
		{	FrameNodeP p=parentFrame;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		{	GraphNodeP p=parentGraph;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		EventPP out=Event::fifo.out;
		EventP ev;
		while (Event::fifo)
		{	Event::fifo >> ev;
			switch (ev->type)
			{
			case Event::ListBoxSelect:
			case Event::ListBoxDClick:
			case Event::ListBoxChar:
			case Event::ListBoxKeyDown:
			case Event::ListBoxKeyUp:
				if (((EventListBox*)ev)->listBox==listBox)
					((EventListBox*)ev)->listBox=0;
				break;
			default:
				;
			}
		}
		Event::fifo.out=out;
		listBox=0;
	}
}
void ListBoxNode::inc_gc() 
{	fifo << onSelect() << onDClick()
		 << onChar() << onKeyDown() << onKeyUp();
}
int ListBoxNode::youngs() 
{	return gmark(onSelect())+gmark(onDClick())
		+gmark(onChar())+gmark(onKeyDown())+gmark(onKeyUp());
}
int ListBoxNode::youngsPtr() 
{	return gmarkPtr(onSelect())+gmarkPtr(onDClick())
		+gmarkPtr(onChar())+gmarkPtr(onKeyDown())+gmarkPtr(onKeyUp());
}
void ListBoxNode::check_oldyoungs() 
{	check_oldyoung(onSelect());
	check_oldyoung(onDClick());
	check_oldyoung(onChar());
	check_oldyoung(onKeyDown());
	check_oldyoung(onKeyUp());
}

wxDataInputStream &ListBoxNode::read(wxDataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	FrameNodeP parentFrameNode;
		GraphNodeP parentGraphNode;
		istr >> (int&)parentFrameNode >> (int&)parentGraphNode >> style;
		listBox=new ListBox_(ListBoxNodeP(this), style);
		listBox->read(istr);
		parentFrame=parentFrameNode;
		parentGraph=parentGraphNode;
	}
	return istr;
}
wxDataOutputStream &ListBoxNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (listBox)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parentFrame) 
			<< GC_::ptrToInt->find(parentGraph)
			<< style;
		listBox->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void ListBoxNode::setPtrs()
{
	parentFrame=(FrameNode*)GC_::intToPtr->find(int(parentFrame));
	parentGraph=(GraphNode*)GC_::intToPtr->find(int(parentGraph));
	if (listBox)
	{	ListBoxP t=listBox;
		setonSelect ((Pair*)GC_::intToPtr->find(int(t->onSelect	)));
		setonDClick ((Pair*)GC_::intToPtr->find(int(t->onDClick	)));
		setonChar   ((Pair*)GC_::intToPtr->find(int(t->onChar   )));
		setonKeyDown((Pair*)GC_::intToPtr->find(int(t->onKeyDown)));
		setonKeyUp  ((Pair*)GC_::intToPtr->find(int(t->onKeyUp  )));
	}
}

void ListBoxNode::recreate()
{
	wxWindow *p=0;
	if (parentFrame)
		p=parentFrame->frame;
	else if (parentGraph)
		p=parentGraph->graph;
	ListBoxP t=listBox, t2=0;
	wxPoint pos=t->GetPosition();
	t2=new ListBox_(ListBoxNodeP(this),
			p, -1, t->choices, 
			t->onSelect, t->onDClick, pos, t->GetSize(), 
			style, wxDefaultValidator, t->GetName());
	listBox=t2;
	wxArrayInt sel;
	int nr=t->listBox->GetSelections(sel);
	for (int i=0; i<nr; i++)
		t2->listBox->SetSelection(sel[i]);
	setonSelect (t->onSelect );
	setonDClick (t->onDClick );
	setonChar   (t->onChar   );
	setonKeyDown(t->onKeyDown);
	setonKeyUp  (t->onKeyUp  );
	t2->listBox->SetForegroundColour(t->listBox->GetForegroundColour());
	t2->listBox->SetBackgroundColour(t->listBox->GetBackgroundColour());
	wxFont f=t->listBox->GetFont();
	t2->listBox->SetFont(f);
	delete t;
	listBox=t2;
}

StringP ListBoxNode::string()
{
	if (listBox)
		return new String(listBox->wxstring());
	return _Null_Word->name();
}

wxString ListBoxNode::wxstring()
{
	if (listBox)
		return listBox->wxstring();
	return wxEmptyString;
}

NodeP lListBox(PairPP arg)
{
	wxWindowID id=-1;
	wxWindow *parent=graph;
	FrameNodeP parentFrameNode=0;
	GraphNodeP parentGraphNode=theApp->mainFrame->graphNode;
	if (arg[1])
	{	switch (arg[1]->type)
		{
		case FRAME:
			parentFrameNode=arg[1]->a_frame(1);
			if throwing return Unbound;
			parent=parentFrameNode->frame;		break;
		
		case GRAPH:
			parentGraphNode=arg[1]->a_graph(1);
			if throwing return Unbound;
			parent=parentGraphNode->graph;		break;

		default:
			error(BAD_DATA, arg[1], 1);
			return Unbound;
		}
	}
	wxString name=_T("ListBox");
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	wxArrayString choices;
	{	PairP list=arg[3]->a_list(3);
		if throwing return Unbound;
		while (list)
		{	StringP s=list->first()->a_string(3); 
			if throwing return Unbound;
			choices.Add(s->ptr());
			next(list);
		}

	}
	PairP onSelect=0;
	if (exist(4))
	{	PairP os=arg[4]->a_runnable(4);
		if throwing return Unbound;
		onSelect=os;
	}
	PairP onDClick=0;
	if (exist(5))
	{	PairP odc=arg[5]->a_runnable(5);
		if throwing return Unbound;
		onDClick=odc;
	}
	long style = 0;
	if (exist(6))
	{	int s=arg[6]->a_int(6);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(7))
	{	PairP p=arg[7]->a_int_vector(7);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(8))
	{	PairP s=arg[8]->a_int_vector(8);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	ListBoxNodeP listBoxNode=new ListBoxNode(
		parentFrameNode, 
		parentGraphNode);
	new ListBox_(listBoxNode,
			parent, id, choices, 
			onSelect, onDClick, 
			pos, size, style, wxDefaultValidator, name);
	return listBoxNode;
}

NodeP lListBoxSelections(PairPP arg)
{
	ListBox_ *listBox=0;
	if (exist(1))
	{	ListBoxNodeP listBoxNode=arg[1]->a_listBox(1);
		if throwing return Unbound;
		listBox=listBoxNode->listBox;
	}
	else
	{	listBox=Event::listBoxValue;
	}
	if (!listBox) return 0;
	wxArrayInt sel;
	int nr=listBox->listBox->GetSelections(sel);
	PairP result=0, last=0, tnode=0;
	for (int i=0; i<nr; i++)
	{	int s=sel[i];
		tnode=new Pair(new Int(s), 0);
		if (!last)
		{	result=tnode;
		}
		else
		{	last->settail(tnode);
		}
		last=tnode;
	}
	return result;
}

NodeP lListBoxSetSelections(PairPP arg)
{
	ListBox_ *listBox=0;
	if (exist(1))
	{	ListBoxNodeP listBoxNode=arg[1]->a_listBox(1);
		if throwing return Unbound;
		listBox=listBoxNode->listBox;
	}
	else
	{	listBox=Event::listBoxValue;
	}
	if (!listBox) return 0;
	PairP list=arg[2]->a_list(2);
	if throwing return Unbound;
	int nr=listBox->listBox->GetCount();
	for (int i=0; i<nr; i++)
		listBox->listBox->Deselect(i);
	while (list)
	{	int sel=list->first()->a_int(2);
		if throwing return Unbound;
		if (sel < 0 || sel >= nr)
		{	error(BAD_DATA, arg[2]);
			return Unbound;
		}
		listBox->listBox->SetSelection(sel);
		next(list);
	}
	return Unbound;
}

NodeP lListBoxSetChoices(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	wxArrayString choices;
	PairP list=arg[2]->a_list(2);
	if throwing return Unbound;
	while (list)
	{	StringP s=list->first()->a_string(2);
		if throwing return Unbound;
		choices.Add(s->ptr());
		next(list);
	}
	listBox->listBox->listBox->Set(choices);
	return Unbound;
}

NodeP lListBoxAppend(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	StringP value=arg[2]->a_string(2);
	if throwing return Unbound;
	listBox->listBox->listBox->Append(value->ptr());
	return Unbound;
}

NodeP lListBoxSetItem(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	int n=arg[2]->a_int(2);
	if throwing return Unbound;
	if (n < 0 || n >= listBox->listBox->listBox->GetCount())
	{	error(BAD_DATA, arg[2]);
		return Unbound;
	}
	StringP str=arg[3]->a_string(3);
	if throwing return Unbound;
	listBox->listBox->listBox->SetString(n, str->ptr());
	return Unbound;
}

NodeP lListBoxRemoveItem(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	int n=arg[2]->a_int(2);
	if throwing return Unbound;
	if (n < 0 || n >= listBox->listBox->listBox->GetCount())
	{	error(BAD_DATA, arg[2]);

		return Unbound;
	}
	listBox->listBox->listBox->Delete(n);
	return Unbound;
}

NodeP lListBoxCount(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	int count=listBox->listBox->listBox->GetCount();
	return new Int(count);
}

#define _TC_ON(NAME)									\
NodeP lListBoxOn##NAME(PairPP arg)			\
{															\
	ListBoxNodeP listBox=arg[1]->a_listBox(1);\
	if throwing return Unbound;					\
	PairP function=arg[2]->a_runnable(2);		\
	if throwing return Unbound;					\
	listBox->seton##NAME(function);			\
	return Unbound;									\
}

_TC_ON(Select)
_TC_ON(DClick)
_TC_ON(Char)
_TC_ON(KeyDown)
_TC_ON(KeyUp)

NodeP lListBoxSetBackgroundColor(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	listBox->listBox->SetBackgroundColour(wc);
	return Unbound;
}

NodeP lListBoxSetColor(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
	listBox->listBox->SetForegroundColour(wc);
	return Unbound;
}

NodeP lListBoxSetFontSize(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=listBox->listBox->GetFont();
	f.SetPointSize(s);
	listBox->listBox->SetFont(f);
	return Unbound;
}

NodeP lListBoxSetFontName(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	StringP s=arg[2]->a_string(2);
	if throwing return Unbound;
	wxFont f=listBox->listBox->GetFont();
	f.SetFaceName(s->ptr());
	listBox->listBox->SetFont(f);
	return Unbound;
}

NodeP lListBoxSetFontStyle(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=listBox->listBox->GetFont();
	f.SetStyle(s);
	listBox->listBox->SetFont(f);
	return Unbound;
}

NodeP lListBoxSetFontWeight(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
	wxFont f=listBox->listBox->GetFont();
	f.SetWeight(s);
	listBox->listBox->SetFont(f);
	return Unbound;
}

NodeP lListBoxEnable(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
	listBox->listBox->Enable(state!=0);
	return Unbound;
}

NodeP lListBoxDestroy(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	{	FrameNodeP p=listBox->parentFrame;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(listBox));
			p->check_oldyoung(p->children);
		}
	}
	{	GraphNodeP p=listBox->parentGraph;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(listBox));
			p->check_oldyoung(p->children);
		}
	}
	listBox->listBox->Destroy();
	return Unbound;
}
